{#
    warning.html
    ~~~~~~~~~~~

    A simple page to emit a warning when no docstring text was found for the
    one a user was looking for

    :copyright: Copyright 2012 by the Spyder team.
    :license: MIT license
#}

<div id="doc-warning">{{text}}</div>
