# -*- coding: utf-8 -*-
#
# Copyright © 2009- The Spyder Development team
# Licensed under the terms of the MIT License
# (see spyderlib/__init__.py for details)

"""
spyderlib.widgets.variableexplorer
==================================

Variable Explorer related widgets
"""
